# -*- encoding: utf-8 -*-

"""
Base API Class Wrapper and Structure for API Integration

The base class is designed to be uniform for all integrated external
data sources with abstract class methods and properties.
"""

import time
import requests
import warnings

from typing import Iterable
from abc import ABC, abstractmethod

# GH#2 disable SSL warnings for self-signed certificates
from urllib3.exceptions import InsecureRequestWarning

class BaseAPI(ABC):
    def __init__(self, apikey : str) -> None:
        self.apikey = apikey

        # create request sessions object for further use
        self._session = requests.Session()


    @property
    @abstractmethod
    def uri(self) -> str:
        pass


    @property
    @abstractmethod
    def headers(self) -> dict:
        pass


    @property
    @abstractmethod
    def endpoint(self) -> str:
        pass


    @property
    @abstractmethod
    def params(self) -> dict:
        pass


    def __format_error__(self, e : requests.exceptions.RequestException) -> str:
        """
        Format the Base Error Response Structure to Hide Sensitive Data

        The make request exposes some sensitive information in the URL
        string, which is not desired for security reasons. Referring
        to an existing issue, the problem is not yet resolved and may
        require an alternate approach.

        Override the method in the child class to preserve the sensitive
        information from leaking, or leave it to the parent class to
        return the value as is.

        :type  e: requests.exceptions.RequestException
        :param e: Exception to be formatted, to hide sensitive
            information from the URL string.
        """

        return e


    def make_request(self, method : str, **kwargs) -> requests.Response:
        sleep = kwargs.get("sleep", 2) # sleep b/w requests, in seconds
        retries = kwargs.get("retries", 3) # no. of retry attempts

        # GH#2 give user the ability to disable SSL verification
        # ? this is not recommended, but can be useful for testing
        verify = kwargs.get("verify", True) # verify SSL certificate

        if _ := kwargs.get("suppresswarning", False):
            warnings.filterwarnings(
                "ignore", category = InsecureRequestWarning
            )
            
            requests.packages.urllib3.disable_warnings(
                InsecureRequestWarning
            )

        # create uri and append the endpoint to the url::
        uri = self.uri + self.endpoint

        # create url load like headers and params
        dataload = {"headers" : self.headers, "params" : self.params}

        # retry on the uri with header and param loads
        for attempt in range(retries):
            try:
                response = self._session.request(
                    method, uri, **dataload, verify = verify
                )
                response.raise_for_status()

                return response # data received, return response

            except requests.exceptions.RequestException as e:
                e = self.__format_error__(e)
                print(f"Request Failed (Attempt: {attempt + 1}) : {e}")

                if attempt == retries - 1:
                    print("Max Retries Reached.")
                    raise e

                time.sleep(sleep)
                continue # continue until retries are exhausted

            except requests.exceptions.Timeout as e:
                print(f"Request Timed Out (Attempt: {attempt + 1}) : {e}")

                if attempt == retries - 1:
                    print("Max Retries Reached.")
                    raise e

                time.sleep(sleep)
                continue # continue until retries are exhausted


    def get(self, **kwargs) -> Iterable:
        """
        Get the Base Response from an External API Source

        The base response is fetched and the data is returned as
        a JSON/Dictionary object to the enduser. The response can be
        further processed and formatted into a desired structure using
        the IO module.
        """

        return self.make_request("GET", **kwargs).json()


    def close(self) -> None:
        self._session.close()



# -*- encoding: utf-8 -*-

"""
A Python Package to Fetch FOREX Rates from API Sources

An one place that integrates different paid/open source API sources
to fetch foreign exchange rates.
"""

# ? package follows https://peps.python.org/pep-0440/
# ? https://python-semver.readthedocs.io/en/latest/advanced/convert-pypi-to-semver.html
__version__ = "v1.1.0"

# init-time options registrations
from forexrates import io
from forexrates import api



# -*- encoding: utf-8 -*-

"""
Operations to Parse Data as :mod:``pandas.DataFrame`` Object
"""

import datetime as dt

from tqdm import tqdm as TQ
from typing import Iterable

import pandas as pd

from forexrates.io.base import BaseIO

class ExchangeRatesIO(BaseIO):
    """
    An IO Parser for the API Response from ExchangeRatesAPI

    The IO parser transforms the API response into a desired format
    by calling any internal methods or processes.

    A single response is received as a JSON (dictionary equivalent)
    from the API and is parsed in a desired format.A typical response
    structure is like:

    .. code-block:: json

        {
            "base" : "EUR",
            "date" : "2020-01-01",
            "rates" : {
                "USD" : 1.2345,
                "INR" : 1.2345,
                [...]
            }
        }

    In case of multiple response parsed under a single call, the
    function is adjusted with parameter to handle such a response. A
    sample response is like:

    .. code-block:: json

        [
            {
                "base" : "EUR",
                "date" : "2020-01-01",
                "rates" : {
                    "USD" : 1.2345,
                    "INR" : 1.2345,
                    [...]
                }
            },
            [...]
        ]

    :type  data: dict
    :param data: A single JSON response from the ExchangeRatesAPI.
        More information about the reponse is available here:
        https://exchangeratesapi.io/documentation/.
    """

    def __init__(self, data : dict | Iterable[dict]) -> None:
        super().__init__(data)


    def dataframe(self, verbose : bool = False, **kwargs) -> pd.DataFrame:
        """
        Parse a JSON (DICTIONARY) Response from ExchangeRatesAPI

        The function provide method to parse a single/multiple
        responses of the module and return a single
        :mod:``pandas.DataFrame`` object.

        :type  verbose: bool
        :param verbose: If set to ``True``, the function will print
            additional information about the parsed dataframe.

        Keyword Arguments
        -----------------
        Currently only setting the column and index names are supported
        using keyword arguments. More control is available in the future
        release and changes over dataframe.

            * **index** (*str*): Name of the index column for parsed
                dataframe. Defaults to ``foreign_exchange_rate``.
            * **column** (*str*): Name of the column for parsed dataframe.
                Defaults to ``target_currency_code``.
            * **basecolumn** (*str*): Name of the base currency column for
                parsed dataframe. Defaults to ``base_currency_code``.
            * **datecolumn** (*str*): Name of the date column for parsed
                dataframe. Defaults to ``effective_date``.


        :rtype:  :mod:``pandas.DataFrame``
        :return: A :mod:``pandas.DataFrame`` object with parsed data.
        """

        index = kwargs.get("index", "foreign_exchange_rate")
        column = kwargs.get("column", "target_currency_code")
        basecolumn = kwargs.get("basecolumn", "base_currency_code")
        datecolumn = kwargs.get("datecolumn", "effective_date")

        frames = []
        for data in TQ(self.data, desc = "Parsing Records..."):
            base = data["base"]
            date = data["date"]

            # the data["date"] is a string in the format "%Y-%m-%d"
            # convert it to a datetime object and return parsed frame
            date = dt.datetime.strptime(date, "%Y-%m-%d")

            frame = pd.DataFrame(
                data["rates"], index = [index]
            ).T.reset_index().rename(columns = {"index" : column})

            frame[basecolumn] = base
            frame[datecolumn] = date

            frames.append(frame[[datecolumn, basecolumn, column, index]])

        frames = pd.concat(frames, ignore_index = True)

        if verbose:
            print(f"Total Parsed Records: {frames.shape[0]:,}")

            min_ = frames[datecolumn].min().date()
            max_ = frames[datecolumn].max().date()
            print(f"  >> Date Range: {min_} - {max_}")

            unique_bases_= frames[basecolumn].unique()
            n_unique_targets_ = frames[column].nunique()
            print(f"  >> Base Currency: {unique_bases_}")
            print(f"  >> # UQ Targets: {n_unique_targets_:,}")

        return frames



# -*- encoding: utf-8 -*-

"""
API Handling for the Exchangerates API (https://exchangeratesapi.io/)

Exchange Rates API provides historic and real-time exchange rates and
currency conversions. More information about the API is available at
https://exchangeratesapi.io/documentation/.
"""

import requests

import datetime as dt
from typing import Iterable
from forexrates.api.base import BaseAPI

class ExchangeRatesAPI(BaseAPI):
    def __init__(self, apikey : str, **kwargs) -> None:
        super().__init__(apikey)

        # ! override default endpoint with endpoint
        # ? the endpoint can be date (%Y-%m-%d) for historic rates
        self._endpoint = kwargs.get("endpoint", "latest")

        # ? each base class accepts parameters as keyword arguments
        self._base = kwargs.get("base", "EUR")


    @property
    def uri(self) -> str:
        return "https://api.exchangeratesapi.io/v1/"
    

    @property
    def endpoint(self) -> str:
        # ? endpoint can be date (%Y-%m-%d) for historic rates
        retval = self._endpoint

        # if date object, then parse as string
        if isinstance(retval, dt.date):
            retval = retval.strftime("%Y-%m-%d")

        return retval
    

    @property
    def params(self) -> dict:
        return {
            "base" : self._base,
            "access_key" : self.apikey
        }
    

    @property
    def headers(self) -> None:
        pass


    def get(self, **kwargs) -> Iterable:
        return super().get(**kwargs)


    def __format_error__(self, e : requests.exceptions.RequestException) -> str:
        server_message = e.response.json()["error"]["message"]
        return f"{e.response} : {e.response.reason} - {server_message}"



# -*- encoding: utf-8 -*-

"""
A Short Module to Handle Input/Output for JSON Response

The response received from the external API sources are parsed and
returned as per user requirements.
"""

from forexrates.io import dataframe



# -*- encoding: utf-8 -*-

"""
Base Class Wrapper and Structure for IO Integration

The base class is designed to be uniform for all integrated external
data sources with abstract class methods and properties.
"""

import pandas as pd

from typing import Iterable
from abc import ABC, abstractmethod

class BaseIO(ABC):
    def __init__(self, data : dict | Iterable[dict]) -> None:
        self.data = self.__parse_data__(data)


    @staticmethod
    def __parse_data__(data : dict | Iterable[dict]) -> pd.DataFrame:
        assert isinstance(data, (dict, list)), \
            "Data must be a dictionary or a list of dictionaries."

        return [data] if isinstance(data, dict) else data


    @abstractmethod
    def dataframe(self, *args, **kwargs) -> pd.DataFrame:
        pass



#!/usr/bin/env python
#
# Copywright (C) 2024 Debmalya Pramanik <neuralNOD@gmail.com>
# LICENSE: MIT License

import os
import sys

from setuptools import setup
from setuptools import find_packages

sys.path.append(
    os.path.join(os.path.dirname(__file__))
)

import forexrates

setup(
    name = "forexrates",
    version = forexrates.__version__,
    author = "shark-utilities developers",
    author_email = "neuralNOD@outlook.com",
    description = "A unified codebase for fetching FOREX rates.",
    long_description = open("README.md", "r").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/sharkutilities/forexrates",
    packages = find_packages(
        exclude = ["tests*", "examples*"]
    ),
    install_requires = [
        "numpy==2.4.2",
        "pandas==3.0.0",
        "psycopg==3.3.2",
        "psycopg-binary==3.3.2",
        "PyYAML==6.0.3",
        "requests==2.32.5",
        "SQLAlchemy==2.0.46",
        "tqdm==4.67.3",
    ],
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "License :: OSI Approved :: MIT License",
    ],
    project_urls = {
        "Issue Tracker" : "https://github.com/sharkutilities/forexrates/issues",
        # "Code Documentations" : "https://.readthedocs.io/en/latest/index.html",
        "Org. Homepage" : "https://github.com/sharkutilities"
    },
    keywords = [
        # keywords for finding the package::
        "forex", "api", "exchange rates", "foreign exchange rates",
        # keywords for finding the package relevant to usecases::
        "wrappers", "data science", "data analysis", "data scientist", "data analyst"
    ],
    python_requires = ">=3.11"
)
